 /*
LABMAN: label manager
loads and interprets the label/doc files
 */
package jcm.gui.doc;

import jcm.core.*;

import java.io.*;
import java.util.*;
import javax.swing.*;
import jcm.core.par.param;
import jcm.core.tls.fileio;
import jcm.core.tls.txt;
import jcm.gui.gen.processdialog;
import static jcm.gui.doc.label.*;
import jcm.gui.doc.label.langcode;
import jcm.gui.doc.label.smd;
import static jcm.core.report.*;
import jcm.gui.nav.jcmAction;

public class labman {
    private static Map<String, label> ldm = new TreeMap();     //Map of the individual labels, sorted by keys
    //use treemap for faster log(n) lookup when hundreds of labels
    static Map<langcode, Set<String>> loaded = new EnumMap(langcode.class);     //Map of categories loaded for each language
    public static Set<String> category = new HashSet();
    static int sy = 1; //syntax type
    public static param<langcode> language = new param("languagemenu", label.langcode.values(), label.langcode.en, complexity.simplest);
    public static boolean recordchanged = false;
    
    static {
        register.setAlwaysOutput(labman.language);
    }
    //*******ACTIONS for Edit Menu***********************************************
    //maybe by keeping the filechooser, wouldn't have to reselect directory each time?
    public static Action loadchangesaction = new jcmAction("Load Changes") {
        public void act() {
            loadFromDirectory(fileio.getFileFromDialog("", "labdocchanges", "Choose Directory for Loading LabDoc Changes", "load"));
        }

    };
    public static Action savechangesaction = new jcmAction("Save Changes") {
        public void act() {
            savedoc(fileio.getFileFromDialog("", "labdocchanges", "Choose Directory for Saving LabDoc Changes", "save"), true);
        }

    };
    public static Action savecopyaction = new jcmAction("Save All") {
        public void act() {
            SwingUtilities.invokeLater(new Runnable() {
                                   public void run() {
                                       loadAllJar();
                                       savedoc(fileio.getFileFromDialog("..", "labdoc", "Choose Directory for Saving LabDoc", "save"), false);
                                   }

                               });
        }

    };    //**** DEFAULT LOAD / SAVE********
    public static void saveonexit() {
        savedoc(fileio.getOutDir("labdocchanges"), true);
    }

    public static void loaddefaultchanges() {
        try { //note checking file exists will give a security exception in "safe mode""
            File f = fileio.getOutDir("labdocchanges");
            if (f.listFiles().length > 0) loadFromDirectory(f);
        //loadAllJar(); //have to load all first before loading changes!
        } catch (Exception e) {
            deb(e);
        }
    }
//******* LOAD*************************
    public static void loadFromDirectory(File f) {
        loadFromJar(langcode.en, ""); //first load english labels, if not done
        String fn = f.getName();
        if (fn.startsWith(".")) return; //ignore mac system files
        if (f.isDirectory()) {
            deb("Updating labels from " + f);
            for (File ff : f.listFiles()) loadFromDirectory(ff);
        } else try {
                int a = fn.indexOf("_"), b = fn.lastIndexOf("_");
                String cat = ((a == b) ? "" : fn.substring(a + 1, b));
                langcode lang = langcode.valueOf(fn.substring(b + 1, b + 3));
                if (lang == null) deb("couldn't identify language of " + fn);
                loadFromJar(lang, cat); //important to load the original from Jar  first (if not already done), since changes will be added onto this
                interpret(fileio.loadstring(f, "UTF8"), lang, cat);
            } catch (Exception e) {
                deb(e, "couldn't load file " + f);
            }
    }
//load ALL labels & doc from existing JAR
    public static void loadAllJarWithDialog() {
        new processdialog("Loading Labels", new JLabel("Please  wait while loading labels"), new Runnable() {
                      public void run() {
                          loadAllJar();
                      }

                  });
    }

    public static void loadAllJar() {
        for (langcode lang : langcode.values()) {
            loadFromJar(lang, "");
            for (String cat : category) loadFromJar(lang, cat);
        }
    }
//LOAD one lang/cat from existing jar
    public static void loadFromJar(langcode lang, String cat) {
        //check if already loaded: note this must be fast, because called in every label look-up!
        if (loaded.get(lang) == null) loaded.put(lang, new HashSet());
        else if (loaded.get(lang).contains(cat)) return; //already loaded it
        loaded.get(lang).add(cat); //won't load again

        String filename = (cat == "" ? "lab" : "doc_" + cat);
        String ls = (lang == langcode.it ? fileio.loadstring("labdoc/" + filename + "_" + lang + ".txt") : fileio.loadstring("labdoc/" + filename + "_" + lang + ".txt", "UTF8"));

        recordchanged = false;
        interpret(ls, lang, cat);
        recordchanged = true;
    }
//**************************************************************
    static void interpret(String ls, langcode lang, String cat) {
        if (ls.length() < 1) return;
        deb("-" + lang + " " + cat);

        //split file by # which is used to separate items
        String[] items = txt.split(ls, "#");

        String key, dat;
        category.add(cat);

        int nt = smd.values().length;
        int[] split = new int[nt + 1];

        for (String item : items) {
            try {
                //identify category headings
                if (item.indexOf("===>") >= 0) {
                    cat = item.substring(item.indexOf("===>") + 4, item.indexOf("<===")).trim();
                    category.add(cat);
                    key = "";
                } else key = convertkey(txt.tonextspace(item, 0));

                label lab = ldm.get(key); //could be more efficient to skip this lookup for the first lab_en file
                if (lab == null) {
                    if (lang != langcode.en) deb(lang + " added " + key); //warning, since key unlikely to be used, if not in english
                    lab = new label(key, cat);
                    ldm.put(key, lab);
                }
                if (lab.cat != cat) lab.cat = cat; //added to cope with category changes

                for (smd t : smd.values()) split[t.ordinal()] = item.indexOf(t.getcode());
                split[nt] = item.length();
                for (int j = nt - 1; j >= 0; j--) {
                    if (split[j] == -1 || split[j] >= split[j + 1]) split[j] = split[j + 1];
                    else {
                        dat = item.substring(split[j] + 1, split[j + 1]).trim();
                        if (j == 0) {
                            if (lang == lang.en) lab.set(lang, smd.i, dat);
                        }  //temporary to remove problem
                        if (j == 1) lab.set(lang, smd.s, dat);
                        if (j == 2) lab.set(lang, smd.m, dat);
                        if (j == 3) lab.set(lang, smd.d, dat);
                        if (j == 4) lab.icon = dat;
                    }
                }
            } catch (Exception e) {
                deb("Problem parsing " + item);
            }
        } //items
        System.gc();
    //deb("loaded labels "+lang+" "+cati);
    }
//***** SAVE *************************************************
//make this anonymous class outside method, otherwise it doesn't load until want to save, and can fail if jar changed
    static Comparator<label> labcomp = new Comparator<label>() {
        public int compare(label a, label b) {
            return a.fo.compareTo(b.fo);
        }

    };
    public static void savedoc(File f, boolean filterchanged) {

        List<label> savelist = new ArrayList(ldm.values());
        Collections.sort(savelist, labcomp);

        Set<langcode> langs = EnumSet.complementOf(EnumSet.of(langcode.sc));

        String oldcat = "dummy", iks, ikso, newcathead;

        Map<langcode, Map<String, StringBuffer>> docsb = new EnumMap(langcode.class);
        for (langcode l : langs) if (l != langcode.sc) {
                docsb.put(l, new HashMap());
                for (String c : category) docsb.get(l).put(c, new StringBuffer());
            }
        Set<langcode> newcat = EnumSet.noneOf(langcode.class);

        //StringBuffer csv=new StringBuffer();

        for (label lab : savelist) {
            //cat=""; for (String c : category) if (c!="" && lab.cat.startsWith(c)) cat=c; //still need startswith?
            //csv.append("\n"+(oldcat!=cat ? cat : "")+"\t"+ntb(lab.key)+"\t"; +ntb(lab.info));

            if (!oldcat.equals(lab.cat)) {
                oldcat = lab.cat;
                for (langcode l : langs) newcat.add(l);
            } //set all to new cat
            newcathead = opencat[sy] + lab.cat + closecat[sy];
            ikso = openkey[sy] + ntb(lab.key) + closekey[sy];

            langloop:
            for (langcode l : langs) {
                iks = ikso + lab.savetext(l, smd.i, sy);
                //added jan09: if deleted, want to keep in labdochanges but not when save all
                if (!filterchanged && lab.isdeleted(l)) continue langloop;
                if ((!filterchanged || lab.getchanged(l, smd.s) != 0 || lab.getchanged(l, smd.m) != 0) && (l == langcode.en || lab.getorig(l, smd.s) != null || lab.getorig(l, smd.m) != null) //note - include if got either sho or med or saving english (since need even for doc lookup)
                        ) {
                    if (newcat.contains(l)) {
                        docsb.get(l).get("").append(newcathead);
                        newcat.remove(l);
                    }
                    docsb.get(l).get("").append("\n" + iks + lab.savetext(l, smd.s, sy) + lab.savetext(l, smd.m, sy) + closeentry[sy]);
                }
                if ((!filterchanged || lab.getchanged(l, smd.d) != 0) && lab.getorig(l, smd.d) != null)
                    docsb.get(l).get(lab.cat).append("\n\n" + iks + lab.savetext(l, smd.d, sy) + closeentry[sy]);
            // note used to apply convertbreaksback()
            //csv.append("\t"+ntb(lab.getorig(l, smd.s))+"\t"+ntb(lab.getorig(l, smd.m))+"\t"+ntb(lab.getorig(l, smd.d)));
            }
        }

        //should also save the changed info!
        //fileio.savetextfile(f, "jcmlabdoctab.txt", csv.toString(), "UTF8");

        if (!f.exists()) f.mkdir();
        for (String c : category) for (langcode l : langs) if (docsb.get(l).get(c).length() > 0) {
                    deb("saving " + l + " " + c);
                    fileio.savetextfile(f, (c == "" ? "lab" : "doc_" + c) + "_" + l + filesuffix[sy], head[sy] + docsb.get(l).get(c).toString() + foot[sy], "UTF8");
                }
    }
//***** SYNTAX ***************************
    static String meta = "<meta http-equiv='Content-Type' content='text/html; charset=UTF-8'/>",  style = "<style> cat { color: orange font: bold 120% } dt { color: red} lab { color: brown } pop { color: green  } link { color: blue } note { font: italic 70% } icon, info { color: magenta; font: italic  } </style>",  extlink = "This is part of the labels / documentation for <a href='http://jcm.chooseclimate.org'>Java Climate Model</a><hr/>";
    static String[] opencat = {"</dl><hr/><dl compact><cat>", "\n\n#! ===>"},  closecat = {"</cat>", "<==="};
    static String[] openkey = {"<dt>", "#"},  closekey = {"</dt><dd> ", ""};
    static String[] closeentry = {"</dd>", ""};
    static String[] filesuffix = {".html", ".txt"};
    static String[] head = {"<html><head>" + meta + style + "</head><body>" + extlink + "<dl  compact>", extlink};
    static String[] foot = {"</dl></html>", ""};//********************************
//FETCH / PUT LABELS
//to get/set the ldm, which is private, ensuring all keys converted to lower-case
    public static boolean gotLabelForKey(String name) {
        return ldm.containsKey(convertkey(name));
    }

    public static void addLabel(String name, label lab) {
        ldm.put(convertkey(name), lab);
    }

    public static label getLabel(String name) {
        return ldm.get(convertkey(name));
    }

    public static List<label> getLabelList() {
        return new ArrayList(ldm.values());
    }

    public static String convertkey(String name) {
        return name.trim().toLowerCase().replace(" ", "_");
    } //key has no spaces, and all lower case

    public static String getShort(String requestkey) {
        return getlabeltext(requestkey, smd.s);
    }

    public static String getTitle(String requestkey) {
        return getlabeltext(requestkey, smd.m);
    }

    public static String getTitleIfDifferent(String requestkey) {
        String popup = getTitle(requestkey);
        if (!popup.equals(getShort(requestkey)) && !popup.equals(getDoc(requestkey))) return "(" + popup + ") ";
        else return "";
    }

    public static String getShortIfDifferent(String requestkey) {
        String popup = getShort(requestkey);
        if (!popup.equals(getTitle(requestkey)) && !popup.equals(getDoc(requestkey))) return "(" + popup + ") ";
        else return "";
    }

    public static String getDoc(String requestkey) {
        if (requestkey.equals("doclist")) return doclist();
        if (requestkey.equals("docsearch")) return docsearch();
        return getlabeltext(requestkey, smd.d);
    }

    public static String getlabeltext(String requestkey, String typecode) {
        for (smd t : smd.values()) if (t.getcode().equals(typecode)) return getlabeltext(requestkey, t);
        return getlabeltext(requestkey, smd.s);
    }

    public static param<langcode> currentlangchooser = language; //pointer to a language param, used by docview
    public static String getlabeltext(String requestkey, smd t) {
        try {
            return getlabeltext(requestkey, t, currentlangchooser.chosen);
        } catch (Exception e) {
            return getlabeltext(requestkey, t, langcode.en);
        } //because the param setup can call getlabel!
    }

    public static String getlabeltext(String requestkey, smd t, langcode lang) {
        String key = convertkey(requestkey);
        try {
            if (lang == langcode.sc && t != smd.d) return key;
            if (key == "") return "";

            //note, this is not very efficient!
            loadFromJar(langcode.en, ""); //first load english labels, if not done
            loadFromJar(lang, ""); //then load labels for specific country (in case labels with & are listed separately in this lang), if not done

            label lab = ldm.get(key);

            //if don't find the key...
            if (lab == null) {
                int split = key.indexOf("&");
                if (split < 0) return ((t != smd.d) ? requestkey.replace("_", " ") : ""); //return requestkey (conserving case)  for short/medium, nothing for doc
                else {	//interpret multi-component labels separated by "&"
                    if (split >= key.length() - 1) return getlabeltext(key.substring(0, split), t); //ignore trailing &
                    String label1 = getlabeltext(key.substring(0, split), t), label2 = getlabeltext(key.substring(split + 1), t);
                    return label1 + (label1.length() > 1 ? " " : "") + label2;
                }
            }

            if (t == smd.d) {
                loadFromJar(langcode.en, lab.cat);
                loadFromJar(lang, lab.cat);
            } //load doc file (if needed) - note need english for default
            return lab.get(lang, t);

        } catch (RuntimeException e) {
            deb(" getlabel " + t + " " + key + " : " + e);
            return ((t != smd.d) ? "!" : "can't find ") + key;
        }
    }

    public static String ntb(String s) {
        return s == null ? "" : s;
    }
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//********************************
//SEARCH
//public static String search(String s) {searchstring=s; return searchresults(); }
    public static void search(String s, boolean a, boolean t) {
        searchstring = s;
        anywords = a;
        titlesonly = t;
    //showwebpage.showinfo(autodoc.makedoc("docsearch"));
    }

    public static String docsearch() {
        return searchresults();
    } //searchbox()+

    public static String searchstring = "";
    public static boolean anywords = false,  titlesonly = false;
    public static String searchresults() {
        if (searchstring.equals("")) return "£§aboutsearch";
        boolean include, found, f;
        float score, score2;
        Map<Float, String> rs = new TreeMap(); //using Treeset, it will be sorted automatically
        String[] ss = txt.split(searchstring.trim(), " ", true);
        String p, q;

        for (label lab : ldm.values()) {
            include = !anywords;
            score = 1;
            try {
                for (int j = 0; j < ss.length; j++) {
                    q = ss[j].toLowerCase();
                    found = false;
                    score2 = 0;

                    Set<langcode> langs = EnumSet.complementOf(EnumSet.of(langcode.sc));
                    for (smd t : smd.values()) if (t != smd.i && !(titlesonly && t == smd.d)) {
                            p = (t != smd.d ? getlabeltext(lab.key, t) : autodoc.parse(getlabeltext(lab.key, t), register.findiob(lab.key), false)).toLowerCase();
                            if (p.length() > 0) for (int r = -1; (r = p.indexOf(q, r + 1)) > 0;) {
                                    found = true;
                                    score2 += 10f / Math.pow(1000 + p.length(), 0.33);
                                }
                        }
                    if (anywords) include |= found;
                    else include &= found;
                    score *= score2;
                }
                if (include) {
                    //<!--"+(1000-(int)score)+"-->("+(score<100 ? " " : "")+(score<10 ? " " : "")
                    rs.put(100 - score, "<li>" + "(" + (int) score + ") @" + lab.key + " " + getTitleIfDifferent(lab.key));
                }
            } catch (RuntimeException e) {
                deb(e, "Search error" + lab.key);
            }
        }

        StringBuffer out = new StringBuffer(" £§aboutsearch ==£`searchresults : " + searchstring + " ==");
        for (String r : rs.values()) out.append(r);
        return out.toString();

    } //search

//******************************************************
//DOCLIST
    public static String doclist() {
        langcode lang = language.chosen;
        loadFromJar(lang, "");
        for (String cat : category) loadFromJar(lang, cat);

        Set<String> rs = new TreeSet(); //automatically sorted (note: used to be sorted by getshortfordoc, looks better)
        String u;
        for (label lab : ldm.values()) {
            try {
                if (!lab.key.startsWith("!") && !lab.key.equals("doclist") && lab.key.length() > 1) {
                    u = " " + (getTitleIfDifferent(lab.key) != getShort(lab.key) ? getTitleIfDifferent(lab.key) + " " : "");
                    rs.add((lab.get(lang.en, smd.d) != null) ? autodoc.link(lab.key) + u : "%%" + getShort(lab.key) + u + "%%");
                }
            } catch (Exception e) {
                deb("doclist problem with key " + lab.key + ": " + e);
            }
        }

        StringBuffer s = new StringBuffer("£§doclistinfo ");
        for (String r : rs) if (r != null) s.append("<li>" + r);
        return s.toString();

    }

} //end class


